"""Module for interacting with binary endpoints."""

import copy
import json
import logging
import os
import re
import struct
from dataclasses import dataclass
from http import HTTPMethod
from io import BytesIO
from pathlib import Path
from tempfile import SpooledTemporaryFile
from typing import IO

import cart
import httpx
import malpz
import pendulum
import tenacity
from azul_bedrock import models_network as azm
from azul_bedrock import models_restapi

from azul_client import exceptions
from azul_client.api.base_api import BaseApiHandler

logger = logging.getLogger(__name__)
DEFAULT_MAX_BYTES_TO_READ = 10 * 1024 * 1024  # 10MB worth of strings.


class _OpenFile:
    """A handler for a potential filepath, raw bytes or a file handle."""

    handle: IO[bytes] | None

    def __init__(self, file_path_or_contents: IO[bytes] | bytes | str | Path):
        self.opened_file = False
        self.file_path_or_contents = file_path_or_contents
        self.handle = None

    def _get_file_handle(self, file_path_or_contents: Path | str | IO[bytes] | bytes):
        if isinstance(file_path_or_contents, bytes):
            self.handle = BytesIO(file_path_or_contents)
        elif isinstance(file_path_or_contents, Path):
            if not file_path_or_contents.exists():
                raise FileExistsError(f"The file with the path {file_path_or_contents=} does not exist.")
            self.opened_file = True
            self.handle = file_path_or_contents.open(mode="rb")
        elif isinstance(file_path_or_contents, str):
            if not os.path.exists(file_path_or_contents):
                raise FileExistsError(f"The file with the path {file_path_or_contents=} does not exist.")
            self.handle = open(file_path_or_contents, mode="rb")
        elif isinstance(file_path_or_contents, SpooledTemporaryFile):
            self.handle = file_path_or_contents
        else:
            # Will already be IO[bytes]
            self.handle = file_path_or_contents

    def open(self) -> IO[bytes]:
        """Open or provide the handle to a file."""
        self._get_file_handle(self.file_path_or_contents)
        return self.handle

    def close(self):
        """If a file was opened close it."""
        if self.opened_file:
            try:
                if self.handle.closed:
                    self.handle.close()
            except Exception:
                print("Failed to close a file.")
                raise

    def __enter__(self):
        """Open or provide the handle to a file."""
        return self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """If a file was opened close it."""
        self.close()


@dataclass
class AugmentedStream:
    """An augmented way of viewing the original binary.

    E.g: the binary is compiled C# code and an augmented stream is a decompiled version of the code.
    E.g2: the original file is a jpg and the augmented stream is a png which has had any potential malware removed.
    """

    label: azm.DataLabel
    file_name: str
    contents_file_path: IO[bytes] | bytes | str | Path | _OpenFile


class _OpenAugmentedStreams:
    opened_streams: list[AugmentedStream]

    def __init__(self, streams: list[AugmentedStream]):
        self.opened_file = False
        self._in_streams = streams
        self.opened_streams = []
        self.file_handles = []
        self.contents = None

    def __enter__(self):
        for stream in self._in_streams:
            # Open the file or contents
            open_file = _OpenFile(stream.contents_file_path)
            self.file_handles.append(open_file)
            # Make a new augmented stream object with the file or contents.
            stream_copy = copy.copy(stream)
            stream_copy.contents_file_path = open_file
            self.opened_streams.append(stream_copy)
        return self.opened_streams

    def __exit__(self, exc_type, exc_val, exc_tb):
        for fh in self.file_handles:
            try:
                fh.close()
            except Exception:
                print("Warning: Failed to close one of the Augmented streams.")


class BinariesData(BaseApiHandler):
    """Interact with binary endpoints."""

    SHA256_regex = r"^[a-fA-F0-9]{64}$"
    upload_download_timeout = 120

    def check_data(self, sha256: str) -> bool:
        """Check data exists for hash."""
        return self._generic_head_request(self.cfg.azul_url + f"/api/v0/binaries/{sha256}/content")

    def download(self, sha256: str) -> bytes:
        """Download binary with the given sha256 in cart format."""
        return self._request(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/content",
            timeout=self.upload_download_timeout,
        ).content

    def download_bulk(self, hashes: list[str]) -> bytes:
        """Download multiple binaries with the given list of sha256 hashes."""
        return self._request(
            method=HTTPMethod.POST,
            url=self.cfg.azul_url + "/api/v0/binaries/content/bulk",
            json={"binaries": hashes},
            timeout=self.upload_download_timeout,
        ).content

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(3),
        wait=tenacity.wait_random(min=1, max=2),
        retry=tenacity.retry_if_exception_type(httpx.TimeoutException),
        before_sleep=tenacity.before_sleep_log(logger, logging.WARNING),
        reraise=True,
    )
    def _base_upload(
        self,
        body: dict,
        *,
        api: str,
        file_path_or_contents: Path | str | bytes | IO[bytes] | SpooledTemporaryFile | None = None,
        augmented_streams: list[AugmentedStream] | None = None,
        filename: str | None = None,
        password: str = "",
        extract: bool = False,
        refresh: bool = False,
    ) -> models_restapi.BinaryData:
        if not augmented_streams:
            augmented_streams = []

        for k, v in list(body.items()):
            if v is None:
                body.pop(k)

        with _OpenFile(file_path_or_contents) as file_handle:
            safe_file = None
            if file_handle is not None:
                # Try to identify this as either a .malpz or .cart
                is_malpz = False
                malpz_header_length = len(malpz.MALPZ_HEADER)
                malpz_header = file_handle.read(malpz_header_length)
                if len(malpz_header) == malpz_header_length:
                    # File too small otherwise
                    try:
                        malpz.validate_version(malpz_header)
                        is_malpz = True
                    except malpz.MetadataException:
                        pass
                file_handle.seek(0)

                cart_header_length = struct.calcsize(cart.MANDATORY_HEADER_FMT)
                cart_header = file_handle.read(cart_header_length)
                is_cart = cart.is_cart(cart_header)
                file_handle.seek(0)

                if not extract and not is_malpz and not is_cart:
                    # This file is not neutered; do this now before sending over the network
                    safe_file = SpooledTemporaryFile(max_size=1000 * 1000)
                    try:
                        print("Packing file as a .CaRT...")
                        cart.pack_stream(file_handle, safe_file)
                        print("CaRT size:", safe_file.tell())
                        safe_file.seek(0)
                    except BaseException:
                        # Avoid leaving temp files in case CaRTing fails. httpx should
                        # handle our file handle otherwise.
                        safe_file.close()
                        raise
                else:
                    # Use the original data source
                    safe_file = file_handle

            with _OpenAugmentedStreams(augmented_streams) as opened_streams:
                stream_data = [
                    ("stream_data", (s.file_name, s.contents_file_path.open(), "application/octet-stream"))
                    for s in opened_streams
                ]
                body["stream_labels"] = [s.label for s in opened_streams]

                main_file = []
                # If there is any contents.
                if safe_file:
                    main_file.append(("binary", (filename, safe_file, "application/octet-stream")))
                resp = self._request_upload(
                    url=self.cfg.azul_url + api,
                    params={"refresh": refresh, "extract": extract, "password": password},
                    files=main_file + stream_data,
                    data=body,
                    timeout=self.upload_download_timeout,
                )

        if resp.status_code != 200 and resp.status_code != 206:
            raise exceptions.bad_response(resp)

        entity = resp.json()[0]
        if not entity.get("id"):
            entity["id"] = entity.get("sha256")
        return models_restapi.BinaryData.model_validate(entity)

    def upload(
        self,
        file_path_or_contents: Path | str | bytes | IO[bytes] | SpooledTemporaryFile,
        source_id: str,
        *,
        references: dict[str, str] | None = None,
        submit_settings: dict[str, str] | None = None,
        augmented_streams: list[AugmentedStream] | None = None,
        filename: str | None = None,
        timestamp: str | None = None,
        security: str,
        extract: bool = False,
        password: str | None = None,
        refresh: bool = False,
        exclude_security_labels: list[str] = None,
        include_queries: bool = False,
    ) -> models_restapi.BinaryData:
        """Upload binary handle with corresponding form data."""
        # If there are no augmented stream and the file isn't being extracted their must be a filename.
        if not augmented_streams and not extract and not filename:
            raise ValueError("If the upload isn't an archive and you aren't uploading streams a filename is required.")

        if not source_id:
            raise ValueError(f"{source_id=} is required to be a valid value.")

        if security is not None and not isinstance(security, str):
            raise ValueError("Security must be a string value.")

        if not timestamp:
            timestamp = pendulum.now(pendulum.UTC).to_iso8601_string()

        references = json.dumps(references) if references else None
        submit_settings = json.dumps(submit_settings) if submit_settings else None

        return self._base_upload(
            body=dict(
                source_id=source_id,
                references=references,
                settings=submit_settings,
                timestamp=timestamp,
                security=security,
                exclude_security_labels=exclude_security_labels,
                include_queries=include_queries,
                filename=filename,
            ),
            api="/api/v0/binaries/source",
            file_path_or_contents=file_path_or_contents,
            augmented_streams=augmented_streams,
            filename=filename,
            extract=extract,
            password=password,
            refresh=refresh,
        )

    def upload_dataless(
        self,
        binary_id: str,  # sha256 of the binary to have it's metadata updated
        source_id: str,
        *,
        references: dict[str, str] | None = None,
        augmented_streams: list[AugmentedStream] | None = None,
        filename: str | None = None,
        timestamp: str | None = None,
        security: str,
        refresh: bool = False,
        exclude_security_labels: list[str] = None,
        include_queries: bool = False,
    ) -> models_restapi.BinaryData:
        """Upload new metadata and potentially alt-streams for a binary."""
        if not binary_id and re.search(self.SHA256_regex, binary_id):
            raise ValueError(f"{binary_id=} must be set to a valid sha256 value.")

        if security is not None and not isinstance(security, str):
            raise ValueError("Security must be a string value.")

        if not timestamp:
            timestamp = pendulum.now(pendulum.UTC).to_iso8601_string()

        references = json.dumps(references) if references else None

        return self._base_upload(
            body=dict(
                sha256=binary_id,
                source_id=source_id,
                references=references,
                timestamp=timestamp,
                security=security,
                exclude_security_labels=exclude_security_labels,
                include_queries=include_queries,
                filename=filename,
            ),
            api="/api/v0/binaries/source/dataless",
            augmented_streams=augmented_streams,
            filename=filename,
            refresh=refresh,
        )

    def upload_child(
        self,
        file_path_or_contents: Path | str | bytes | IO[bytes],
        parent_sha256: str,
        relationship: dict[str, str],
        *,
        submit_settings: dict[str, str] | None = None,
        parent_type: str = "binary",
        filename: str | None = None,
        timestamp: str | None = None,
        security: str,
        extract: bool = False,
        password: str | None = None,
        refresh: bool = False,
        exclude_security_labels: list[str] = None,
        include_queries: bool = False,
    ) -> models_restapi.BinaryData:
        """Upload a child binary and attach it to the parent binary with the provided sha256 ID."""
        if not parent_sha256 or not re.search(self.SHA256_regex, parent_sha256):
            raise ValueError(f"{parent_sha256=} must be set to a valid sha256 value.")

        if not relationship:
            raise ValueError(f"{relationship=} must be a dictionary with at least one key value pair.")
        relationship = json.dumps(relationship) if relationship else None
        submit_settings = json.dumps(submit_settings) if submit_settings else None

        if not extract and not filename:
            raise ValueError("If the upload isn't an archive a filename is required.")

        if security is not None and not isinstance(security, str):
            raise ValueError("Security must be a string value.")

        if not timestamp:
            timestamp = pendulum.now(pendulum.UTC).to_iso8601_string()

        return self._base_upload(
            body=dict(
                timestamp=timestamp,
                security=security,
                exclude_security_labels=exclude_security_labels,
                include_queries=include_queries,
                relationship=relationship,
                settings=submit_settings,
                parent_type=parent_type,
                parent_sha256=parent_sha256,
                filename=filename,
            ),
            api="/api/v0/binaries/child",
            file_path_or_contents=file_path_or_contents,
            filename=filename,
            extract=extract,
            password=password,
            refresh=refresh,
        )

    def expedite_processing(self, sha256: str, *, bypass_cache: bool = False) -> None:
        """Expedite or reprocess the file with the provided sha256.

        If bypass_cache is on ensure plugins actually re-process the binary and don't rely on the cache.
        """
        self._request(
            method=HTTPMethod.POST,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/expedite",
            params={"bypass_cache": bypass_cache},
        )

    def download_augmented_stream(self, sha256: str, stream_sha256: str):
        """Download the raw augmented stream for a given submission binary.

        First sha256 is the sha256 of the binary and the second sha256 is that of the augmented stream.
        """
        return self._request(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/content/{stream_sha256}",
        ).content

    def download_hex(
        self, sha256, *, offset: int = 0, max_bytes_to_read: int | None = None, shortform: bool = False
    ) -> models_restapi.BinaryHexView:
        """Download either all or a section of the raw hex of a file.

        Typically used in chunks to download sections of a file until either you have the whole file or have
        enough information from the file.

        :param str sha256: sha256 of the file you want to download hex for.
        :param int offset: starting offset for where to download the hex from.
        :param int max_bytes_to_read: bytes to read before stopping and returning what you have.
        :param bool shortform: If true, will return 16 hex bytes as a string instead of 16 strings in a list.
        """
        params = {"offset": offset, "max_bytes_to_read": max_bytes_to_read, "shortform": shortform}
        params = self.filter_none_values(params)

        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/hexview",
            response_model=models_restapi.BinaryHexView,
            params=params,
        )

    def get_strings(
        self,
        sha256: str,
        *,
        min_length: int = 4,
        max_length: int = 200,
        offset: int = 0,
        max_bytes_to_read: int = DEFAULT_MAX_BYTES_TO_READ,
        take_n_strings: int = 1000,
        filter: str | None = None,
        regex: str | None = None,
        file_format: str | None = None,
    ) -> models_restapi.BinaryStrings:
        """Get strings for a binary file with multiple potential additional parameters.

        :param str sha256: File to get strings for.
        :param int min_length: Minimum length of string (when decoded).
        :param int max_length: Maximum length of string (when decoded).
        :param int offset: Search for strings from offset.
        :param int max_bytes_to_read: How many bytes to search for, default of 10MB.
        :param int take_n_strings: ow many strings to return.
        :param str filter: Case-insensitive search string to filter strings with.
        :param str regex: Regex pattern to search strings with.
        :param str file_format: Optional file type for AI string filter.
        """
        params = {
            "min_length": min_length,
            "max_length": max_length,
            "offset": offset,
            "max_bytes_to_read": max_bytes_to_read,
            "take_n_strings": take_n_strings,
            "filter": filter,
            "regex": regex,
            "file_format": file_format,
        }
        params = self.filter_none_values(params)

        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/strings",
            response_model=models_restapi.BinaryStrings,
            params=params,
        )

    def search_hex(
        self,
        sha256: str,
        filter: str,
        *,
        offset: int = 0,
        max_bytes_to_read: int | None = None,
        take_n_hits: int = 1000,
    ) -> models_restapi.BinaryStrings:
        """Search a hex file with the given filter.

        :param str sha256: Search the data file that has this sha256.
        :param str filter: Search a hex file with the given hex string filter.
        :param int offset: Search for hits from offset.
        :param int max_bytes_to_read: How many bytes to search for, if this is not set, return to EOF.
        :param int take_n_hits: Maximum number of hits to return.
        """
        params = {
            "offset": offset,
            "max_bytes_to_read": max_bytes_to_read,
            "take_n_hits": take_n_hits,
            "filter": filter,
        }
        params = self.filter_none_values(params)

        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/search/hex",
            response_model=models_restapi.BinaryStrings,
            params=params,
        )



from azul_bedrock import models_restapi

from .base_test import BaseApiTest


class TestStatisticsApi(BaseApiTest):
    def test_get_statistics(self):
        """Verify UserAccess is being acquired correctly from the API."""
        resp = self.api.statistics.get_statistics()
        self.assertGreaterEqual(resp.binary_count, 2)

    def test_meta_is_set_appropriately(self):
        """Verify metadata is set to a value for stats which should set it."""
        self.api.statistics.get_statistics()
        meta = self.api.statistics.get_meta_from_last_request()
        self.assertIsInstance(meta, models_restapi.Meta)



from azul_bedrock import models_restapi
from azul_bedrock import models_settings as azs

from azul_client.exceptions import BadResponse404

from .base_test import BaseApiTest


class TestSourcesApi(BaseApiTest):
    def test_get_all_sources(self):
        """Verify sources exist and are in the appropriate form."""
        dict_of_sources = self.api.sources.get_all_sources()
        print(dict_of_sources)
        self.assertIsInstance(dict_of_sources, dict)
        source_name_list = list(dict_of_sources.keys())
        # Should be 3 or more sources.
        self.assertIn(self.known_source, source_name_list)
        self.assertGreater(len(source_name_list), 3)
        self.assertIsInstance(source_name_list[0], str)

        self.assertIsInstance(dict_of_sources[self.known_source], azs.Source)
        # Verify the description is not the default meaning it was set to something.
        self.assertGreater(len(dict_of_sources[self.known_source].description), 2)

    def test_get_meta_for_all_sources(self):
        """Verify meta data is gathered correctly."""
        self.api.sources.get_all_sources()
        meta = self.api.sources.get_meta_from_last_request()
        self.assertIsInstance(meta, models_restapi.Meta)
        # Verify we have some kind of security string
        self.assertGreater(len(meta.security), 2)

    def test_check_source_exists(self):
        """Verify that the requested sources either exist or don't."""
        does_sort_exist = self.api.sources.check_source_exists(self.known_source)
        self.assertTrue(does_sort_exist)

        # Verify a random source name doesn't exist
        does_sort_exist = self.api.sources.check_source_exists("JSDLKFJVIEURNDCKdmf")
        self.assertFalse(does_sort_exist)

    def test_read_source(self):
        source_info = self.api.sources.read_source(self.known_source)
        self.assertIsInstance(source_info, models_restapi.Source)
        # Verify there is some data in the source.
        self.assertGreaterEqual(source_info.num_entities, 2)
        self.assertEqual(source_info.name, self.known_source)

    def test_read_source_references(self):
        source_ref = self.api.sources.read_source_references(self.known_source)
        self.assertIsInstance(source_ref, models_restapi.References)
        self.assertGreater(len(source_ref.items), 1)

        with self.assertRaises(BadResponse404):
            self.api.sources.read_source_references(self.known_source, term="termquery")

        source_ref = self.api.sources.read_source_references(self.known_source, term=self.test_username)
        self.assertIsInstance(source_ref, models_restapi.References)
        self.assertGreaterEqual(len(source_ref.items), 1)



import time

import pendulum
from azul_bedrock import models_network as azm
from azul_bedrock import models_restapi
from azul_bedrock.models_restapi import binaries_auto_complete as bedr_bauto

from azul_client.exceptions import BadResponse, BadResponse404

from .base_test import BaseApiTest


class TestBinaryMetaOther(BaseApiTest):
    @classmethod
    def setUpClass(cls):
        cls.non_existent_sha256 = "ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f74"
        cls.tag_to_create = "test-tag"
        super().setUpClass()

    def test_verify_sha256s_length(self):
        """Ensure the sha256's are the appropriate length.

        This covers the tests that loop over sha256 to make sure their assertions are actually getting called.
        For the actual length refer to the module setup in __init__
        """
        self.assertGreaterEqual(len(self.sha256s), 10)

    def test_check_meta(self):
        for sha in self.sha256s:
            exists = self.api.binaries_meta.check_meta(sha)
            self.assertEqual(exists, True)

        exists = self.api.binaries_meta.check_meta(self.non_existent_sha256)
        self.assertEqual(exists, False)

    def test_get_meta(self):
        first = self.sha256s[0]
        meta = self.api.binaries_meta.get_meta(first, bucket_size=1000)

        source_ids: list[str] = []
        references: list[dict] = []
        for source in meta.sources:
            for d_source in source.direct:
                source_ids.append(d_source.name)
                references.append(d_source.references)

        self.assertIn(self.known_source, source_ids)
        self.assertIn(self.upload_ref, references)

        # test non-existent file
        self.assertRaises(BadResponse404, self.api.binaries_meta.get_meta, self.non_existent_sha256)

        self.assertIn(self.known_source, source_ids)
        self.assertIn(self.upload_ref, references)

        meta_less = self.api.binaries_meta.get_meta(
            first, details=[models_restapi.BinaryMetadataDetail.documents], bucket_size=1000
        )
        self.assertEqual(0, len(meta_less.sources))

        # The detailed version of the binaries metadata should include sufficient additional information
        # to be at least x characters longer than the original metadata
        simple = meta_less.model_dump_json(exclude_unset=True, exclude_defaults=True)
        complex = meta.model_dump_json(exclude_unset=True, exclude_defaults=True)
        self.assertLess(len(simple) + 200, len(complex))

    def test_get_model(self):
        model = self.api.binaries_meta.get_model()
        self.assertIsInstance(model, models_restapi.EntityModel)
        # There should be some keys.
        self.assertGreater(len(model.keys), 1)

    def test_find_autocomplete(self):
        # No data yet
        resp = self.api.binaries_meta.find_autocomplete("")
        print(resp)
        self.assertEqual(resp.type, "Initial")
        self.assertIsInstance(resp, bedr_bauto.AutocompleteInitial)

        # Results
        resp = self.api.binaries_meta.find_autocomplete("a")
        print(resp)
        self.assertEqual(resp.type, "FieldValue")
        self.assertIsInstance(resp, bedr_bauto.AutocompleteFieldValue)
        resp = self.api.binaries_meta.find_autocomplete("b")
        print(resp)
        self.assertEqual(resp.type, "FieldValue")
        self.assertIsInstance(resp, bedr_bauto.AutocompleteFieldValue)

        # No results
        resp = self.api.binaries_meta.find_autocomplete("kjhasdfhjkkjhdfskjhasfdkjh")
        self.assertEqual(resp.type, "FieldValue")
        self.assertIsInstance(resp, bedr_bauto.AutocompleteFieldValue)

        # Trailing Space causing None return type.
        resp = self.api.binaries_meta.find_autocomplete("dnksjvdkjndvs  ")
        self.assertEqual(resp.type, "None")
        self.assertIsInstance(resp, bedr_bauto.AutocompleteNone)
        print(resp)

        # Invalid input for auto-completion
        resp = self.api.binaries_meta.find_autocomplete("$@!!@#$%^&*()")
        print(resp)
        self.assertEqual(resp.type, "Error")
        self.assertIsInstance(resp, bedr_bauto.AutocompleteError)

    def test_get_has_newer_metadata(self):
        # Check if there has been any new data an hour in the future (this should be impossible)
        now = pendulum.now()
        now_plus_an_hour = now + pendulum.duration(hours=1)
        resp = self.api.binaries_meta.get_has_newer_metadata(self.sha256s[0], now_plus_an_hour.isoformat())
        self.assertEqual(resp.count, 0)

        # Check if there has been any new data ever. (there should be)
        forever_ago = pendulum.datetime(2000, 1, 1, 1, 1, 1, 1, tz=pendulum.UTC)
        resp = self.api.binaries_meta.get_has_newer_metadata(self.sha256s[0], forever_ago.isoformat())
        # verify at least one new piece of data.
        self.assertGreater(resp.count, 0)

    def test_get_similar_ssdeep_entities(self):
        # Empty ssdeep should create a value error
        with self.assertRaises(ValueError):
            self.api.binaries_meta.get_similar_ssdeep_entities("")

        # Invalid ssdeep should cause an API exception
        with self.assertRaises(BadResponse):
            self.api.binaries_meta.get_similar_ssdeep_entities("hkjdsfhjksdf")

        first = self.sha256s[0]
        binary_list = self.api.binaries_meta.find_hashes(hashes=[first])
        self.assertEqual(len(binary_list.items), 1)
        first_binary_data = binary_list.items[0]
        first_binary_ssdeep = first_binary_data.ssdeep
        resp = self.api.binaries_meta.get_similar_ssdeep_entities(ssdeep=first_binary_ssdeep)
        # Should be at least a couple of matches because we generate a lot of very similar files.
        self.assertGreater(len(resp.matches), 4)

    def test_get_similar_tlsh_entities(self):
        # Empty tlsh should create a value error
        with self.assertRaises(ValueError):
            self.api.binaries_meta.get_similar_tlsh_entities("")

        # Invalid tlsh (v9 doesn't exist) should cause an API exception
        with self.assertRaises(BadResponse):
            self.api.binaries_meta.get_similar_tlsh_entities("T9jdsfhjksdf")

        first = self.sha256s[0]
        binary_list = self.api.binaries_meta.find_hashes(hashes=[first])
        self.assertEqual(len(binary_list.items), 1)
        first_binary_data = binary_list.items[0]
        first_binary_tlsh = first_binary_data.tlsh
        resp = self.api.binaries_meta.get_similar_tlsh_entities(tlsh=first_binary_tlsh)
        self.assertGreater(len(resp.matches), 1)

    def test_get_similar_entities(self):
        first = self.sha256s[0]
        similar_entities = self.api.binaries_meta.get_similar_entities(first)
        self.assertIsInstance(similar_entities, models_restapi.SimilarMatch)

    def test_get_nearby_entities(self):
        first = self.sha256s[0]
        nearby_resp = self.api.binaries_meta.get_nearby_entities(first)
        self.assertIsInstance(nearby_resp, models_restapi.ReadNearby)
        self.assertEqual(nearby_resp.id_focus, first)
        # There should at least be 1 links (the node)
        self.assertGreaterEqual(len(nearby_resp.links), 1)

        nearby_resp = self.api.binaries_meta.get_nearby_entities(
            first, include_cousins=models_restapi.IncludeCousinsEnum.Large
        )
        self.assertIsInstance(nearby_resp, models_restapi.ReadNearby)
        self.assertEqual(nearby_resp.id_focus, first)
        # There should at least be 1 links (the node)
        self.assertGreaterEqual(len(nearby_resp.links), 1)

        nearby_resp = self.api.binaries_meta.get_nearby_entities(
            first, include_cousins=models_restapi.IncludeCousinsEnum.No
        )
        self.assertIsInstance(nearby_resp, models_restapi.ReadNearby)
        self.assertEqual(nearby_resp.id_focus, first)
        # There should at least be 1 links (the node)
        self.assertGreaterEqual(len(nearby_resp.links), 1)

        nearby_resp = self.api.binaries_meta.get_nearby_entities(self.non_existent_sha256)
        # No links should be present on non-existing binary.
        self.assertEqual(nearby_resp.id_focus, self.non_existent_sha256)
        self.assertEqual(len(nearby_resp.links), 0)

    def test_get_binary_tags(self):
        """Test getting tags works (also tested in create and delete tag methods)."""
        first = self.sha256s[0]
        tags = self.api.binaries_meta.get_binary_tags(first)
        self.assertIsInstance(tags, models_restapi.ReadAllEntityTags)

        # Non-existent hash has nothing to get.
        resp = self.api.binaries_meta.get_binary_tags(self.non_existent_sha256)
        self.assertEqual(len(resp.items), 0)

    def test_create_and_delete_tag_on_binary(self):
        """Tests a tag can be seen with get and then creates and deletes the tag.

        If the tag already exists, it is first deleted and then created.
        """
        # Maximum number of times to recheck a tag exists before giving up.
        max_check_retry_count = 5

        def check_tag_exists(sha256: str) -> bool:
            """Check a tag exists on a binary."""
            found_tags = self.api.binaries_meta.get_binary_tags(sha256)
            for cur_tag in found_tags.items:
                if cur_tag.tag == self.tag_to_create:
                    return True
            return False

        def verify_delete(sha256: str):
            """Delete a tag and verify it's no longer present on the binary."""
            self.api.binaries_meta.delete_tag_on_binary(sha256, self.tag_to_create)
            is_tag_exists = check_tag_exists(sha256)
            retry_count = 0
            while is_tag_exists and retry_count < max_check_retry_count:
                retry_count += 1
                time.sleep(1)
                is_tag_exists = check_tag_exists(sha256)
            return is_tag_exists

        def verify_create(sha256: str):
            """Create a tag and verify it's on the binary."""
            self.api.binaries_meta.create_tag_on_binary(sha256, self.tag_to_create, self.default_security)
            is_tag_exists = check_tag_exists(sha256)
            retry_count = 0
            while not is_tag_exists and retry_count < max_check_retry_count:
                retry_count += 1
                time.sleep(1)
                is_tag_exists = check_tag_exists(sha256)
            return is_tag_exists

        first = self.sha256s[0]
        # Check if tag is already present and then delete or create the tag first accordingly.
        is_tag_already_present = check_tag_exists(first)

        # Iteration 1 delete or create.
        if is_tag_already_present:
            self.assertFalse(verify_delete(first))
        else:
            self.assertTrue(verify_create(first))

        # Iteration 2 delete or create.
        if is_tag_already_present:
            self.assertTrue(verify_create(first))
        else:
            self.assertFalse(verify_delete(first))

        # Create on non existent sha256 (no error nothing happens)
        self.api.binaries_meta.create_tag_on_binary(
            self.non_existent_sha256, self.tag_to_create, self.default_security
        )
        # Delete on non existent sha256 (no error nothing happens)
        self.api.binaries_meta.delete_tag_on_binary(self.non_existent_sha256, self.tag_to_create)

    def test_get_binary_status(self):
        # Verify some plugins have status events.
        first = self.sha256s[0]
        binary_statuses = self.api.binaries_meta.get_binary_status(first)
        self.assertGreater(len(binary_statuses.items), 1)
        self.assertIsInstance(binary_statuses.items[0], models_restapi.StatusEvent)

        # No events if it doesn't exist.
        with self.assertRaises(BadResponse404):
            binary_statuses = self.api.binaries_meta.get_binary_status(self.non_existent_sha256)

    def test_get_binary_documents(self):
        # Standard loading of documents.
        first = self.sha256s[0]
        opensearch_docs = self.api.binaries_meta.get_binary_documents(first)
        self.assertGreaterEqual(len(opensearch_docs.items), 1)
        # Find at least one sourced event
        opensearch_docs = self.api.binaries_meta.get_binary_documents(first, action=azm.BinaryAction.Sourced, size=200)
        self.assertGreaterEqual(len(opensearch_docs.items), 1)

        # Non existent binary shouldn't have any sourced events.
        with self.assertRaises(BadResponse404):  # Expecting 404
            self.api.binaries_meta.get_binary_documents(self.non_existent_sha256, action=azm.BinaryAction.Sourced)



import copy
import io
import unittest
from tempfile import SpooledTemporaryFile

from azul_bedrock import models_network as azm

from azul_client import Api
from azul_client.api import binaries_data
from azul_client.config import get_config
from azul_client.exceptions import BadResponse, BadResponse404

from . import module_ref, module_sha256s, module_source


class BaseUploadTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.api = Api(get_config())
        cls.source = module_source
        cls.ref = module_ref
        cls.classifications = ["OFFICIAL//TLP:GREEN"]
        cls.non_existent_sha256 = "ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f74"
        cls.submit_settings = {"test": "test_setting"}


class TestBinaryUpload(BaseUploadTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def setUp(self):
        self.std_content = io.BytesIO(b"content")
        self.std_kwargs = {
            "source_id": self.source,
            "filename": "test_basic_upload.txt",
            "security": "OFFICIAL",
            "references": self.ref,
        }

    def basic_response_check(self, entity: azm.BinaryEvent.Entity):
        self.assertEqual(entity.sha256, "ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f73")
        self.assertEqual(entity.file_format, "text/plain")
        self.assertEqual(entity.mime, "text/plain")

    def test_basic_upload(self):
        resp = self.api.binaries_data.upload(self.std_content, **self.std_kwargs)
        self.basic_response_check(resp)

    def test_basic_upload_with_settings(self):
        resp = self.api.binaries_data.upload(self.std_content, **self.std_kwargs, submit_settings=self.submit_settings)
        self.basic_response_check(resp)

    def test_basic_upload_spooled_file(self):
        with SpooledTemporaryFile(max_size=100000) as spooledFile:
            spooledFile.write(self.std_content.read())
            spooledFile.seek(0)
            resp = self.api.binaries_data.upload(spooledFile, **self.std_kwargs)
            self.basic_response_check(resp)

    def test_upload_with_security(self):
        self.std_kwargs["security"] = self.classifications[0]
        resp = self.api.binaries_data.upload(self.std_content, **self.std_kwargs)
        self.basic_response_check(resp)

    def test_upload_with_aug(self):
        resp = self.api.binaries_data.upload(
            self.std_content,
            **self.std_kwargs,
            augmented_streams=[
                binaries_data.AugmentedStream(
                    label=azm.DataLabel.TEST, file_name="file1.exe", contents_file_path=io.BytesIO(b"type1")
                ),
                binaries_data.AugmentedStream(
                    label=azm.DataLabel.ASSEMBLYLINE, file_name="file2.exe", contents_file_path=io.BytesIO(b"type2")
                ),
                binaries_data.AugmentedStream(
                    label=azm.DataLabel.TEXT, file_name="file3.exe", contents_file_path=io.BytesIO(b"type3")
                ),
            ],
        )
        self.basic_response_check(resp)

    def test_invalid_source_upload(self):
        self.std_kwargs["source_id"] = "random_source_id_that_doesnt_exist"
        self.assertRaises(
            BadResponse,
            self.api.binaries_data.upload,
            self.std_content,
            **self.std_kwargs,
        )


class TestBinaryUploadDataless(BaseUploadTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.parent_sha256 = module_sha256s[0]

    def setUp(self):
        self.std_kwargs = {
            "security": "OFFICIAL",
            "references": self.ref,
        }
        self.non_existent_sha256 = "ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f74"

    def basic_response_check(self, entity: azm.BinaryEvent.Entity):
        self.assertEqual(entity.sha256, self.parent_sha256)
        self.assertEqual(entity.file_format, "text/plain")
        self.assertEqual(entity.mime, "text/plain")

    def test_basic_upload(self):
        resp = self.api.binaries_data.upload_dataless(self.parent_sha256, self.source, **self.std_kwargs)
        self.basic_response_check(resp)

    def test_upload_with_security(self):
        self.std_kwargs["security"] = self.classifications[0]
        resp = self.api.binaries_data.upload_dataless(
            self.parent_sha256, self.source, filename="newfilename.txt", **self.std_kwargs
        )
        self.basic_response_check(resp)

    def test_upload_with_aug(self):
        resp = self.api.binaries_data.upload_dataless(
            self.parent_sha256,
            self.source,
            augmented_streams=[
                binaries_data.AugmentedStream(
                    label=azm.DataLabel.TEXT, file_name="file1.exe", contents_file_path=io.BytesIO(b"type1")
                ),
                binaries_data.AugmentedStream(
                    label=azm.DataLabel.TEST, file_name="file2.exe", contents_file_path=io.BytesIO(b"type2")
                ),
            ],
            **self.std_kwargs,
        )
        self.basic_response_check(resp)

    def test_invalid_source_upload(self):
        self.assertRaises(
            BadResponse,
            self.api.binaries_data.upload_dataless,
            self.parent_sha256,
            "random_source_id_that_doesnt_exist",
            **self.std_kwargs,
        )

    def test_invalid_parent_sha256(self):
        self.assertRaises(
            BadResponse404,
            self.api.binaries_data.upload_dataless,
            self.non_existent_sha256,  # in theory this hash isn't in the system
            self.source,
            **self.std_kwargs,
        )


class TestBinaryUploadChild(BaseUploadTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.parent_sha256s = copy.copy(module_sha256s)

    def setUp(self):
        self.std_kwargs = {"security": "OFFICIAL", "filename": "child-file.txt"}
        self.relationship = {"test-parent": "parent-in-test"}
        self.child_content = io.BytesIO(b"Child-content")
        self.parent_sha256 = self.parent_sha256s.pop()

    def basic_response_check(self, entity: azm.BinaryEvent.Entity):
        self.assertEqual(entity.sha256, "d2dba47e53f9d97060b3e5fee9416d06177aee5e2e9488f8643d1c3f55cc9f6b")
        self.assertEqual(entity.file_format, "text/plain")
        self.assertEqual(entity.mime, "text/plain")

    def test_basic_upload(self):
        resp = self.api.binaries_data.upload_child(
            self.child_content,
            self.parent_sha256,
            self.relationship,
            **self.std_kwargs,
            submit_settings=self.submit_settings,
        )
        self.basic_response_check(resp)

    def test_upload_with_security(self):
        self.std_kwargs["security"] = self.classifications[0]
        resp = self.api.binaries_data.upload_child(
            self.child_content, self.parent_sha256, self.relationship, **self.std_kwargs
        )
        self.basic_response_check(resp)

    def test_no_parent_child_relationship(self):
        self.assertRaises(
            ValueError,
            self.api.binaries_data.upload_child,
            self.child_content,
            self.parent_sha256,
            {},
            **self.std_kwargs,
        )

    def test_invalid_parent_sha256(self):
        self.assertRaises(
            BadResponse,
            self.api.binaries_data.upload_child,
            self.child_content,
            self.non_existent_sha256,
            self.relationship,
            **self.std_kwargs,
        )



"""Base test case used to pass common variables and test functionalities to all test cases."""

import copy
import unittest

from azul_client.api import Api
from azul_client.config import get_config

from . import module_ref, module_sha256s, module_source, upload_security


class BaseApiTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.api = Api(get_config())
        # A known source that should always be there
        cls.known_source = module_source
        # Sha256's
        cls.sha256s = copy.copy(module_sha256s)
        cls.upload_ref = module_ref
        # Name of a plugin unlikely to ever be removed
        cls.entropy_plugin_name = "Entropy"
        # Name of a plugin unlikely to ever be removed that is python based.
        cls.python_known_plugin_name = "Alphabets"
        # Name of the test user.
        cls.test_username = module_ref.get("user")
        # Default level of security to assign to new objects during testing.
        cls.default_security = upload_security
        cls.max_security = "MORE OFFICIAL"
        cls.min_security = "LESS OFFICIAL"

        cls.non_existent_sha256 = "ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f74"
        super().setUpClass()



from azul_bedrock import models_restapi
from azul_bedrock.models_auth import UserInfo

from .base_test import BaseApiTest


class TestUsersApi(BaseApiTest):
    def setUp(self):
        super().setUp()

        # Ensure this is unset between runs
        self.api.set_excluded_security(None)

    def test_get_security_settings(self):
        """Verify UserAccess is being acquired correctly from the API."""
        opensearch_user_access = self.api.users.get_opensearch_user_info()
        self.assertIsInstance(opensearch_user_access, models_restapi.UserAccess)
        print(opensearch_user_access)
        # Verify a role is in opensearch that is unlikely to change.
        self.assertIn("azul_read", opensearch_user_access.roles)

    def test_meta_is_none_appropriately(self):
        """Verify metadata is set to none for a query that doesn't set it."""
        self.api.users.get_opensearch_user_info()
        meta = self.api.sources.get_meta_from_last_request()
        self.assertIsNone(meta)

    def test_get_user_infos(self):
        """Verify UserInfo is being acquired correctly from the API."""
        user_info = self.api.users.get_user_info()
        print(user_info)
        self.assertIsInstance(user_info, UserInfo)
        # Verify a common user_info role is present.
        self.assertIn("azul_read", user_info.roles)

    def test_get_security_settings_with_excluded_security(self):
        """Test excluded security doesn't cause an issue on an API endpoint that doesn't use it."""
        for class_to_exclude in [self.min_security, self.max_security, self.default_security]:
            self.api.set_excluded_security([class_to_exclude])
            opensearch_user_access = self.api.users.get_opensearch_user_info()
            self.assertIsInstance(opensearch_user_access, models_restapi.UserAccess)
            # Verify a role is in opensearch that is unlikely to change.
            self.assertIn("azul_read", opensearch_user_access.roles)

    def test_get_user_infos_with_excluded_security(self):
        """Test exclude security works when it's meant to exclude security information."""
        for class_to_exclude in [self.min_security, self.max_security, self.default_security]:
            self.api.set_excluded_security([class_to_exclude])
            user_info = self.api.users.get_user_info()
            self.assertIsInstance(user_info, UserInfo)
            # Verify a common user_info role is present.
            self.assertIn("azul_read", user_info.roles)



import json
import os
import re
import sys
import tempfile
from unittest import mock

import pytest
from azul_bedrock import models_restapi
from click.testing import CliRunner
from pytest_httpx import HTTPXMock

from azul_client import Api, client
from azul_client import config as mconfig

from . import dummy_entity, dummy_entity_meta

test_file = os.path.join(os.path.dirname(__file__), "data", "testfile.txt")
test_folder = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture
def api() -> Api:
    """Fixture container the initialised api."""
    mconfig.location = ""
    cur_api = Api(
        mconfig.Config(
            auth_type="none",
            oidc_url="http://localhost",
            auth_client_id="servicer",
        )
    )
    client.api = cur_api
    return cur_api


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def mod_client(api: Api):
    client.api = api


def test_get(api: Api, runner: CliRunner, httpx_mock: HTTPXMock):
    """Just testing that it runs, not its output."""
    httpx_mock.add_response(method="POST", json={"data": {"items": []}})
    runner.invoke(
        client.get,
        [],
        catch_exceptions=False,
    )


def test_get_meta(api: Api, runner: CliRunner, httpx_mock: HTTPXMock, dummy_entity_meta: dict):
    """Tests that the get-meta command correctly returns formatted JSON for a given binary."""
    httpx_mock.add_response(
        method="GET",
        url=f"{api.config.azul_url}/api/v0/binaries/randomsha256?bucket_size=100",
        json={"data": dummy_entity_meta},
    )

    res = runner.invoke(
        client.get_meta,
        "randomsha256",
    )
    print(res.stdout)
    print(res.exception)
    assert res.exit_code == 0

    # Assert that the response is valid JSON (as this shouldn't be an interactive terminal)
    assert "security" in json.loads(res.stdout)


def test_get_meta_pretty(api: Api, runner: CliRunner, httpx_mock: HTTPXMock, dummy_entity_meta: dict):
    """Tests that pretty printing a binary works correctly."""
    httpx_mock.add_response(
        method="GET",
        url=f"{api.config.azul_url}/api/v0/binaries/randomsha256?bucket_size=100",
        json={"data": dummy_entity_meta},
    )

    # Check that forcing pretty printing returns a Pydantic repr - this won't be paginated, but this isn't
    # reasonable to test here (as it is OS dependent)
    res = runner.invoke(client.get_meta, ["randomsha256", "--pretty"])
    print(res.stdout)
    print(res.exception)
    assert res.exit_code == 0

    assert res.stdout.startswith("BinaryMetadata(")


def test_get_meta_file(api: Api, runner: CliRunner, httpx_mock: HTTPXMock, dummy_entity_meta: dict):
    """Tests that file metadata can be written to a file."""
    httpx_mock.add_response(
        method="GET",
        url=f"{api.config.azul_url}/api/v0/binaries/randomsha256?bucket_size=100",
        json={"data": dummy_entity_meta},
    )

    # Check that content can be written to a file
    with tempfile.NamedTemporaryFile() as file:
        res = runner.invoke(client.get_meta, ["randomsha256", "--output", file.name])
        print(res)
        print(res.stdout)
        print(res.stderr)
        print(res.exception)
        assert res.exit_code == 0
        assert "saving output to path %s\n" % file.name == res.stderr

        assert "security" in json.loads(file.read())


@pytest.mark.httpx_mock(can_send_already_matched_responses=True)
def test_put(api: Api, runner: CliRunner, httpx_mock: HTTPXMock, dummy_entity: dict):
    with mock.patch("click.confirm") as c:
        c.return_value = "y"

        # we are just testing that it runs, not its output
        httpx_mock.add_response(
            url=re.compile(rf".*\/api\/v0\/binaries\/source\?.*"),
            method="POST",
            json=[dummy_entity],
        )
        # test folder
        res = runner.invoke(
            client.put,
            [test_folder, "testing", "--security", "OFFICIAL"],
        )
        print(res.stdout)
        print(res.exception)
        assert res.exit_code == 0
        assert (
            re.compile("2 files found including:\n/testfile.txt\n/deep/testfile2.txt").search(res.stdout) is not None
        )
        assert "\nSource: testing" in res.stdout
        assert "\nReferences: {}" in res.stdout
        assert "\nSecurity: OFFICIAL" in res.stdout

        # test basic
        res = runner.invoke(
            client.put,
            [test_file, "testing", "--security", "OFFICIAL"],
        )
        print(res.stdout)
        print(res.exception)
        assert res.exit_code == 0
        assert re.compile("1 files found including:\ntestfile.txt\n").search(res.stdout) is not None
        assert "\nSource: testing" in res.stdout
        assert "\nReferences: {}" in res.stdout
        assert "\nSecurity: OFFICIAL" in res.stdout

        # test with references
        res = runner.invoke(
            client.put,
            [test_file, "testing", "--ref", "user:llama", "--ref", "group:primary", "--security", "OFFICIAL"],
        )
        print(res.stdout)
        print(res.exception)
        assert res.exit_code == 0
        assert re.compile("1 files found including:\ntestfile.txt\n").search(res.stdout) is not None
        assert "\nSource: testing" in res.stdout
        assert "\nReferences: {'user': 'llama', 'group': 'primary'}" in res.stdout
        assert "\nSecurity: OFFICIAL" in res.stdout

        # test with security
        res = runner.invoke(
            client.put,
            [test_file, "testing", "--security", "OFFICIAL"],
        )
        print(res.stdout)
        print(res.exception)
        assert res.exit_code == 0
        assert re.compile("1 files found including:\ntestfile.txt\n").search(res.stdout) is not None
        assert "\nSource: testing" in res.stdout
        assert "\nReferences: {}" in res.stdout
        assert "\nSecurity: OFFICIAL" in res.stdout

        # test with timestamp
        res = runner.invoke(
            client.put,
            [test_file, "testing", "--timestamp", "20220101T000000", "--security", "OFFICIAL"],
        )
        print(res.stdout)
        print(res.exception)
        assert res.exit_code == 0
        assert re.compile("1 files found including:\ntestfile.txt\n").search(res.stdout) is not None
        assert "\nSource: testing" in res.stdout
        assert "\nReferences: {}" in res.stdout
        assert "\nTimestamp: 2022-01-01T00:00:00Z" in res.stdout
        assert "\nSecurity: OFFICIAL" in res.stdout


def _read_test_file_to_std_in():
    with open(test_file, "rb") as f:
        sys.stdin = f.read()


@pytest.mark.httpx_mock(can_send_already_matched_responses=True)
def test_put_stdin(api: Api, runner: CliRunner, httpx_mock: HTTPXMock, dummy_entity: dict):
    # we are just testing that it runs, not its output
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/source\?.*"),
        method="POST",
        json=[dummy_entity],
    )
    _read_test_file_to_std_in()
    # test basic
    res = runner.invoke(
        client.put_stdin,
        ["Spooled-test-file1.txt", "testing", "--security", "OFFICIAL", "-y"],
    )
    print(res.stdout)
    print(res.exception)
    assert res.exit_code == 0
    assert "Filename: Spooled-test-file1.txt" in res.stdout
    assert "\nSource: testing" in res.stdout
    assert "\nReferences: {}" in res.stdout
    assert "\nSecurity: OFFICIAL" in res.stdout

    _read_test_file_to_std_in()
    # test with references
    res = runner.invoke(
        client.put_stdin,
        [
            "Spooled-test-file1.txt",
            "testing",
            "--ref",
            "user:llama",
            "--ref",
            "group:primary",
            "--security",
            "OFFICIAL",
            "-y",
        ],
    )
    print(res.stdout)
    print(res.exception)
    assert res.exit_code == 0
    assert "Filename: Spooled-test-file1.txt" in res.stdout
    assert "\nSource: testing" in res.stdout
    assert "\nReferences: {'user': 'llama', 'group': 'primary'}" in res.stdout
    assert "\nSecurity: OFFICIAL" in res.stdout

    _read_test_file_to_std_in()
    # test with security
    res = runner.invoke(
        client.put_stdin,
        [
            "Spooled-test-file1.txt",
            "testing",
            "--security",
            "OFFICIAL",
            "-y",
        ],
    )
    print(res.stdout)
    print(res.exception)
    assert res.exit_code == 0
    assert "Filename: Spooled-test-file1.txt" in res.stdout
    assert "\nSource: testing" in res.stdout
    assert "\nReferences: {}" in res.stdout
    assert "\nSecurity: OFFICIAL" in res.stdout

    _read_test_file_to_std_in()
    # test with timestamp
    res = runner.invoke(
        client.put_stdin,
        [
            "Spooled-test-file1.txt",
            "testing",
            "--timestamp",
            "20220101T000000",
            "--security",
            "OFFICIAL",
            "-y",
        ],
    )
    print(res.stdout)
    print(res.exception)
    assert res.exit_code == 0
    assert "Filename: Spooled-test-file1.txt" in res.stdout
    assert "\nSource: testing" in res.stdout
    assert "\nReferences: {}" in res.stdout
    assert "\nTimestamp: 2022-01-01T00:00:00Z" in res.stdout
    assert "\nSecurity: OFFICIAL" in res.stdout


@pytest.mark.httpx_mock(can_send_already_matched_responses=True)
def test_put_child(api: Api, runner: CliRunner, httpx_mock: HTTPXMock, dummy_entity: dict):
    with mock.patch("click.confirm") as c:
        c.return_value = "y"

        # we are just testing that it runs, not its output
        httpx_mock.add_response(
            url=re.compile(rf".*\/api\/v0\/binaries\/child\?.*"),
            method="POST",
            json=[dummy_entity],
        )

        # Test base case with folder
        res = runner.invoke(
            client.put_child,
            [
                test_folder,
                "--parent",
                "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "--relationship",
                "action:transformed",
                "--security",
                "OFFICIAL",
            ],
        )
        assert res.exit_code == 0
        assert (
            re.compile("2 files found including:\n/testfile.txt\n/deep/testfile2.txt").search(res.stdout) is not None
        )
        assert "\nParent: e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855" in res.stdout
        assert "\nRelationship: {'action': 'transformed'}" in res.stdout
        assert "\nSecurity: OFFICIAL" in res.stdout

        # Test base case with a valid file
        res = runner.invoke(
            client.put_child,
            [
                test_file,
                "--parent",
                "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "--relationship",
                "action:transformed",
                "--security",
                "OFFICIAL",
            ],
        )
        assert res.exit_code == 0
        assert re.compile("1 files found including:\ntestfile.txt\n").search(res.stdout) is not None
        assert "\nParent: e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855" in res.stdout
        assert "\nRelationship: {'action': 'transformed'}" in res.stdout
        assert "\nSecurity: OFFICIAL" in res.stdout

        # Test multiple references
        res = runner.invoke(
            client.put_child,
            [
                test_file,
                "--parent",
                "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "--relationship",
                "action:transformed",
                "--relationship",
                "cool:action",
                "--security",
                "OFFICIAL",
            ],
        )
        assert res.exit_code == 0
        assert re.compile("1 files found including:\ntestfile.txt\n").search(res.stdout) is not None
        assert "\nParent: e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855" in res.stdout
        assert "\nRelationship: {'action': 'transformed', 'cool': 'action'}" in res.stdout
        assert "\nSecurity: OFFICIAL" in res.stdout

        # Test timestamp
        res = runner.invoke(
            client.put_child,
            [
                test_file,
                "--parent",
                "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "--relationship",
                "action:transformed",
                "--security",
                "OFFICIAL",
                "--timestamp",
                "20220101T000000",
            ],
        )
        assert res.exit_code == 0
        assert re.compile("1 files found including:\ntestfile.txt\n").search(res.stdout) is not None
        assert "\nParent: e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855" in res.stdout
        assert "\nRelationship: {'action': 'transformed'}" in res.stdout
        assert "\nTimestamp: 2022-01-01T00:00:00Z" in res.stdout
        assert "\nSecurity: OFFICIAL" in res.stdout

        # Test invalid sha256
        res = runner.invoke(
            client.put_child,
            [
                test_file,
                "--parent",
                "notsha256",
                "--relationship",
                "action:transformed",
                "--security",
                "OFFICIAL",
            ],
        )
        assert res.exit_code == 1


def _setup_find_mock(api: Api, httpx_mock: HTTPXMock):
    httpx_mock.reset()
    httpx_mock.add_response(
        url=re.compile(rf"{api.config.azul_url}/api/v0/binaries.*"),
        method="POST",
        status_code=200,
        json={
            "data": models_restapi.EntityFind(
                items_count=1,
                items=[
                    models_restapi.EntityFindItem(
                        key="abc123",
                        exists=True,
                        has_content=True,
                        sha256="abc123",
                    )
                ],
            ).model_dump()
        },
    )


def _make_find_assertions(
    httpx_mock: HTTPXMock,
    term: str = [],
    max_entities: int = [],
    sort_prop: models_restapi.FindBinariesSortEnum = [],
    sort_asc: bool = [],
):
    req = httpx_mock.get_requests()
    assert req[0].url.params.get_list("term") == term
    assert req[0].url.params.get_list("max_entities") == max_entities
    assert req[0].url.params.get_list("sort") == sort_prop
    assert req[0].url.params.get_list("sort_asc") == sort_asc


def test_get(api: Api, runner: CliRunner, httpx_mock: HTTPXMock, dummy_entity: dict):
    # Test base case simple query with all options
    _setup_find_mock(api, httpx_mock)

    res = runner.invoke(
        client.get,
        ["--term", "source.name:testing"],
    )
    assert res.exit_code == 0, res.stdout
    _make_find_assertions(httpx_mock, ["source.name:testing"], max_entities=[str(100)], sort_asc=["false"])

    # Test output directory
    _setup_find_mock(api, httpx_mock)
    "/api/v0/binaries/{sha256}/content"
    httpx_mock.add_response(
        url=f"{api.config.azul_url}/api/v0/binaries/abc123/content",
        method="GET",
        status_code=200,
        content=b"not-a-real-file",
    )

    with tempfile.TemporaryDirectory() as cur_dir:
        number_of_files_originally = len(os.listdir(cur_dir))
        res = runner.invoke(
            client.get,
            ["--output", cur_dir, "--term", "source.name:testing"],
        )
        assert res.exit_code == 0, res.stdout
        number_of_files_after = len(os.listdir(cur_dir))
        assert number_of_files_originally + 1 == number_of_files_after
    _make_find_assertions(httpx_mock, ["source.name:testing"], max_entities=[str(100)], sort_asc=["false"])

    httpx_mock.reset()
    # Test all options.
    _setup_find_mock(api, httpx_mock)

    res = runner.invoke(
        client.get,
        [
            "--term",
            "source.name:testing",
            "--max",
            "10",
            "--sort-by",
            str(models_restapi.FindBinariesSortEnum.timestamp),
            "--sort-asc",
        ],
    )
    assert res.exit_code == 0, res.stdout
    _make_find_assertions(
        httpx_mock,
        ["source.name:testing"],
        max_entities=[str(10)],
        sort_prop=[models_restapi.FindBinariesSortEnum.timestamp],
        sort_asc=["true"],
    )


def test_download_normal(api: Api, runner: CliRunner, httpx_mock: HTTPXMock):
    """Test the basic download case."""
    # Download a single file.
    dummy_hash_a = "a" * 64
    httpx_mock.add_response(
        url=f"{api.config.azul_url}/api/v0/binaries/{dummy_hash_a}/content",
        method="GET",
        status_code=200,
        content=b"not-a-real-file-a",
    )

    with tempfile.TemporaryDirectory() as cur_dir:
        number_of_files_originally = len(os.listdir(cur_dir))
        res = runner.invoke(
            client.download,
            [dummy_hash_a, "--output", cur_dir],
        )
        assert res.exit_code == 0, res.stdout
        number_of_files_after = len(os.listdir(cur_dir))
        assert number_of_files_originally + 1 == number_of_files_after


def test_download_multiple(api: Api, runner: CliRunner, httpx_mock: HTTPXMock):
    """Test downloading two hashes at the same time."""
    # Test downloading two files.
    dummy_hash_a = "a" * 64
    dummy_hash_b = "b" * 64
    httpx_mock.add_response(
        url=f"{api.config.azul_url}/api/v0/binaries/{dummy_hash_a}/content",
        method="GET",
        status_code=200,
        content=b"not-a-real-file-a",
    )
    httpx_mock.add_response(
        url=f"{api.config.azul_url}/api/v0/binaries/{dummy_hash_b}/content",
        method="GET",
        status_code=200,
        content=b"not-a-real-file-b",
    )

    with tempfile.TemporaryDirectory() as cur_dir:
        number_of_files_originally = len(os.listdir(cur_dir))
        res = runner.invoke(
            client.download,
            [dummy_hash_a, dummy_hash_b, "--output", cur_dir],
        )
        assert res.exit_code == 0, res.stdout
        number_of_files_after = len(os.listdir(cur_dir))
        print(os.listdir(cur_dir))
        assert number_of_files_originally + 2 == number_of_files_after


def test_download_multiple_with_bad(api: Api, runner: CliRunner, httpx_mock: HTTPXMock):
    """Test downloading two files with one bad sha256 provided and one missing sha256 provided."""
    dummy_hash_a = "a" * 64
    dummy_hash_b = "b" * 64
    # hash not in azul
    dummy_hash_c = "c" * 64
    httpx_mock.add_response(
        url=f"{api.config.azul_url}/api/v0/binaries/{dummy_hash_a}/content",
        method="GET",
        status_code=200,
        content=b"not-a-real-file-a",
    )
    httpx_mock.add_response(
        url=f"{api.config.azul_url}/api/v0/binaries/{dummy_hash_b}/content",
        method="GET",
        status_code=200,
        content=b"not-a-real-file-b",
    )
    httpx_mock.add_response(
        url=f"{api.config.azul_url}/api/v0/binaries/{dummy_hash_c}/content",
        method="GET",
        status_code=404,
    )

    with tempfile.TemporaryDirectory() as cur_dir:
        number_of_files_originally = len(os.listdir(cur_dir))
        res = runner.invoke(
            client.download,
            [dummy_hash_a, "not-a-sha256", dummy_hash_b, dummy_hash_c, "--output", cur_dir],
        )
        assert res.exit_code == 0, res.stdout
        number_of_files_after = len(os.listdir(cur_dir))
        assert number_of_files_originally + 2 == number_of_files_after



"""Mappings for the Azul features API."""

import logging
from http import HTTPMethod

from azul_bedrock import models_restapi
from pydantic import TypeAdapter

from azul_client.api.base_api import BaseApiHandler

logger = logging.getLogger(__name__)

feature_type_count_adapter = TypeAdapter(dict[str, models_restapi.FeatureMulticountRet])
entities_in_features_count_adapter = TypeAdapter(dict[str, dict[str, models_restapi.ValueCountRet]])
entities_in_featurevalueparts_count_adapter = TypeAdapter(dict[str, dict[str, models_restapi.ValuePartCountRet]])


class Features(BaseApiHandler):
    """API for counting, and requesting features and their values from Azul."""

    def count_unique_values_in_feature(
        self, items: list[str], *, skip_count: bool = False, author: str = "", author_version: str = ""
    ) -> dict[str, models_restapi.FeatureMulticountRet]:
        """Count number of unique values for provided feature(s)."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/features/values/counts",
            method=HTTPMethod.POST,
            response_model=feature_type_count_adapter,
            params={
                "skip_count": skip_count,
                "author": author,
                "author_version": author_version,
            },
            json={"items": items},
            get_data_only=True,
        )

    def count_unique_entities_in_features(
        self, items: list[str], *, skip_count: bool = False, author: str = "", author_version: str = ""
    ) -> dict[str, models_restapi.FeatureMulticountRet]:
        """Count number of unique entities for provided feature(s)."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/features/entities/counts",
            method=HTTPMethod.POST,
            response_model=feature_type_count_adapter,
            params={
                "skip_count": skip_count,
                "author": author,
                "author_version": author_version,
            },
            json={"items": items},
            get_data_only=True,
        )

    def count_unique_entities_in_featurevalues(
        self,
        items: list[models_restapi.ValueCountItem],
        *,
        skip_count: bool = False,
        author: str = "",
        author_version: str = "",
    ) -> dict[str, dict[str, models_restapi.ValueCountRet]]:
        """Count unique entities for multiple feature values."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/features/values/entities/counts",
            method=HTTPMethod.POST,
            response_model=entities_in_features_count_adapter,
            params={
                "skip_count": skip_count,
                "author": author,
                "author_version": author_version,
            },
            json={"items": [x.model_dump() for x in items]},
            get_data_only=True,
        )

    def count_unique_entities_in_featurevalueparts(
        self,
        items: list[models_restapi.ValuePartCountItem],
        *,
        skip_count: bool = False,
        author: str = "",
        author_version: str = "",
    ) -> dict[str, dict[str, models_restapi.ValuePartCountRet]]:
        """Count unique entities for multiple value parts."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/features/values/parts/entities/counts",
            method=HTTPMethod.POST,
            response_model=entities_in_featurevalueparts_count_adapter,
            params={
                "skip_count": skip_count,
                "author": author,
                "author_version": author_version,
            },
            json={"items": [x.model_dump() for x in items]},
            get_data_only=True,
        )

    def get_all_feature_value_tags(self) -> models_restapi.ReadFeatureValueTags:
        """Get a list and count of all feature value tags."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/features/all/tags",
            method=HTTPMethod.GET,
            response_model=models_restapi.ReadFeatureValueTags,
            get_data_only=True,
        )

    def get_feature_values_in_tag(self, tag: str) -> models_restapi.ReadFeatureTagValues:
        """Get feature values that are tagged with the provided tag."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + f"/api/v0/features/tags/{tag}",
            method=HTTPMethod.GET,
            response_model=models_restapi.ReadFeatureTagValues,
            get_data_only=True,
        )

    def create_feature_value_tag(self, tag: str, feature: str, value: str, security: str) -> bool:
        """Create a feature value tag."""
        return self._request(
            method=HTTPMethod.POST,
            url=self.cfg.azul_url + f"/api/v0/features/tags/{tag}",
            params={"feature": feature, "value": value},
            json={"security": security},
        ).json()

    def delete_feature_value_tag(self, tag: str, feature: str, value: str) -> bool:
        """Delete a feature value tag."""
        return self._request(
            method=HTTPMethod.DELETE,
            url=self.cfg.azul_url + f"/api/v0/features/tags/{tag}",
            params={"feature": feature, "value": value},
        ).json()

    def find_features(self, *, author: str = "", author_version: str = "") -> models_restapi.Features:
        """Find all features optionally selecting a specific author and author_version to search for."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/features",
            method=HTTPMethod.GET,
            response_model=models_restapi.Features,
            params={"author": author, "author_version": author_version},
            get_data_only=True,
        )

    def find_values_in_feature(
        self,
        feature: str,
        *,
        term: str = "",
        sort_asc: bool = True,
        case_insensitive: bool = False,
        author: str = "",
        author_version: str = "",
        num_values: int = 500,
        after: str = "",
    ) -> models_restapi.ReadFeatureValues:
        """Find all the values associated with a feature.

        Note - After should only ever be set when you want to paginate and you get the after value from the previous
        query response.
        """
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + f"/api/v0/features/feature/{feature}",
            method=HTTPMethod.POST,
            response_model=models_restapi.ReadFeatureValues,
            params={
                "term": term,
                "sort_asc": sort_asc,
                "case_insensitive": case_insensitive,
                "author": author,
                "author_version": author_version,
                "num_values": num_values,
            },
            json={"after": after},
            get_data_only=True,
        )



"""Module for interacting with security endpoints."""

from http import HTTPMethod

from azul_client.api.base_api import BaseApiHandler


class Security(BaseApiHandler):
    """Interact with binary endpoints."""

    def get_security_settings(self) -> dict:
        """Get the current Azul security settings."""
        return self._request(method=HTTPMethod.GET, url=self.cfg.azul_url + "/api/v0/security").json()

    def normalise(self, security: str) -> str:
        """Validates and normalises a security string."""
        return self._request(
            method=HTTPMethod.POST, url=self.cfg.azul_url + "/api/v1/security/normalise", json={"security": security}
        ).json()

    def get_max_security_string(self, security_strings: list[str]) -> str:
        """Get the max security from the provided security strings."""
        return self._request(
            method=HTTPMethod.POST, url=self.cfg.azul_url + "/api/v1/security/max", json=security_strings
        ).json()

    def get_is_user_an_admin(self) -> bool:
        """Get a boolean value indicating if the authenticated user is an admin or not."""
        return self._request(method=HTTPMethod.GET, url=self.cfg.azul_url + "/api/v0/security/is_admin").json()



"""Mappings for the Azul statistics API."""

import logging
from http import HTTPMethod

from azul_bedrock import models_restapi

from azul_client.api.base_api import BaseApiHandler

logger = logging.getLogger(__name__)


class Statistics(BaseApiHandler):
    """API for accessing statistics about Azul from the Azul restapi."""

    def get_statistics(self) -> models_restapi.StatisticSummary:
        """Get statistics about the current state of azul."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/statistics",
            method=HTTPMethod.GET,
            response_model=models_restapi.StatisticSummary,
            get_data_only=True,
        )



from unittest import TestCase

from azul_client.api.binaries_meta import FindOptions


class BinaryFindUnitTest(TestCase):
    def test_find_with_zero_integer_cases(self):
        """Regression test to ensure setting an int value to 0 still includes the filter."""
        find_options = FindOptions(source_depth=[0])
        self.assertEqual("depth:0", find_options.to_query())

        find_options = FindOptions(source_depth=0)
        self.assertEqual("depth:0", find_options.to_query())

        find_options = FindOptions(source_depth=None)
        self.assertEqual("", find_options.to_query())

        find_options = FindOptions(source_depth=[])
        self.assertEqual("", find_options.to_query())

        find_options = FindOptions(greater_than_size_bytes=0)
        self.assertEqual("size:>0", find_options.to_query())

        find_options = FindOptions(greater_than_size_bytes=None)
        self.assertEqual("", find_options.to_query())

    def test_find_with_empty_string(self):
        """Regression test to allow for empty strings to be set for values."""
        find_options = FindOptions(sources="")
        self.assertEqual('source.name:""', find_options.to_query())

        find_options = FindOptions(sources=[""])
        self.assertEqual('source.name:""', find_options.to_query())

        find_options = FindOptions(sources="abcdef")
        self.assertEqual('source.name:"abcdef"', find_options.to_query())

        find_options = FindOptions(sources=None)
        self.assertEqual("", find_options.to_query())



import io
import json
import re

import cart
import malpz
import pytest
import python_multipart
from azul_bedrock import models_network as azm
from azul_bedrock import models_restapi
from cart.cart import CART_MAGIC, TRAC_MAGIC
from pytest_httpx import HTTPXMock

from azul_client import Api
from azul_client.api import binaries_data
from azul_client.config import Config


@pytest.fixture
def api() -> Api:
    """Fixture container the initialised api."""
    return Api(
        Config(
            auth_type="none",
            azul_url="http://localhost",
        )
    )


@pytest.fixture
def dummy_entity() -> dict:
    return azm.BinaryEvent.Entity(sha256="random").model_dump()


@pytest.fixture
def dummy_entity_meta() -> dict:
    """Taken from output of the integration test."""
    return models_restapi.BinaryMetadata(
        documents=models_restapi.BinaryDocuments(count=250, newest="2024-05-10T00:28:57.749Z"),
        security=["OFFICIAL"],
        sources=[
            models_restapi.BinarySource(
                source="testing",
                direct=[
                    models_restapi.EventSource(
                        security="OFFICIAL",
                        name="testing",
                        timestamp="2024-05-09T04:26:03.174062Z",
                        references={"user": "azul-client-test"},
                        track_source_references="testing.985fad029e65f5f68005505c8f6b57ba",
                    )
                ],
                indirect=[],
            )
        ],
        tags=[],
        parents=[],
        instances=[
            models_restapi.EntityInstance(
                key="plugin.ExifTool.binary_enriched.",
                author=models_restapi.EntityInstanceAuthor(
                    security="OFFICIAL", category="plugin", name="ExifTool", version="2024.04.29"
                ),
                action=azm.BinaryAction.Enriched,
                stream=None,
                num_feature_values=7,
            )
        ],
        features=[],
        streams=[],
        info=[
            {
                "info": {
                    "entropy": {"blocks": [], "overall": 3.4548223999466066, "block_count": 0, "block_size": 256}
                },
                "instance": "plugin.entropy.binary_enriched.",
            }
        ],
    ).model_dump()


def test_upload(api: Api, httpx_mock: HTTPXMock, dummy_entity: dict):
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/source\?.*"),
        method="POST",
        status_code=200,
        json=[dummy_entity],
    )
    api.binaries_data.upload(
        io.BytesIO(b"content"),
        source_id="source1",
        security="",
        augmented_streams=[
            binaries_data.AugmentedStream(
                label="data1", file_name="file1.exe", contents_file_path=io.BytesIO(b"type1")
            ),
            binaries_data.AugmentedStream(
                label="data2", file_name="file2.exe", contents_file_path=io.BytesIO(b"type2")
            ),
            binaries_data.AugmentedStream(
                label="data3", file_name="file3.exe", contents_file_path=io.BytesIO(b"type3")
            ),
        ],
    )
    last_request = httpx_mock.get_request()
    body = last_request.read()
    assert b'Content-Disposition: form-data; name="stream_data"; filename="file1.exe"\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_data"; filename="file2.exe"\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_data"; filename="file3.exe"\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_labels"\r\n\r\ndata1\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_labels"\r\n\r\ndata2\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_labels"\r\n\r\ndata3\r\n' in body
    assert b'Content-Disposition: form-data; name="security"\r\n\r\n\r\n' in body

    # test security
    httpx_mock.reset()
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/source\?.*"), method="POST", status_code=200, json=[dummy_entity]
    )
    api.binaries_data.upload(
        io.BytesIO(b"content"),
        "source1",
        security="OFFICIAL",
        filename="dummy",
    )
    last_request = httpx_mock.get_request()
    body = last_request.read()
    print(body.decode("utf-8", errors="replace"))
    assert b'Content-Disposition: form-data; name="security"\r\n\r\nOFFICIAL\r\n' in body

    # more security tests
    httpx_mock.reset()
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/source\?.*"), method="POST", status_code=200, json=[dummy_entity]
    )
    api.binaries_data.upload(
        io.BytesIO(b"content"),
        "source1",
        security="OFFICIAL//TLP:GREEN",
        filename="dummy",
    )
    body = httpx_mock.get_request().read()
    print(body.decode("utf-8", errors="replace"))
    assert b'Content-Disposition: form-data; name="security"\r\n\r\nOFFICIAL//TLP:GREEN\r\n' in body

    # test no filename
    pytest.raises(ValueError, api.binaries_data.upload, io.BytesIO(b"content"), "source1", security="OFFICIAL")

    # test bad source id
    pytest.raises(
        ValueError, api.binaries_data.upload, io.BytesIO(b"content"), "", security="OFFICIAL", filename="abc.txt"
    )


def test_upload_neutering(api: Api, httpx_mock: HTTPXMock, dummy_entity: dict):
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/source\?.*"),
        method="POST",
        status_code=200,
        json=[dummy_entity],
    )

    # Unneutered file should be CaRT'd
    api.binaries_data.upload(
        io.BytesIO(b"content"),
        filename="testfile",
        source_id="source1",
        security="",
        augmented_streams=[],
    )
    last_request = httpx_mock.get_request()
    body = last_request.read()
    assert CART_MAGIC in body
    assert TRAC_MAGIC in body

    # CaRT'd files should be not be CaRT'd again
    httpx_mock.reset()
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/source\?.*"),
        method="POST",
        status_code=200,
        json=[dummy_entity],
    )

    buffer = io.BytesIO()
    cart.pack_stream(io.BytesIO(b"content"), buffer, optional_header="EXTRA CART CONTENT")
    buffer.seek(0)
    api.binaries_data.upload(
        buffer,
        filename="testfile",
        source_id="source1",
        security="",
        augmented_streams=[],
    )
    last_request = httpx_mock.get_request()
    body = last_request.read()

    assert CART_MAGIC in body
    assert TRAC_MAGIC in body

    # Validate that the file itself is only CaRT'd once
    file_contents: python_multipart.multipart.File | None = None

    def field_handler(_field):
        pass

    def file_handler(file):
        nonlocal file_contents
        file_contents = file

    python_multipart.parse_form(last_request.headers, io.BytesIO(body), field_handler, file_handler)

    handle = file_contents.file_object  # type: ignore
    handle.seek(0)

    output_contents = io.BytesIO()
    header, _ = cart.unpack_stream(handle, output_contents)

    assert b"content" == output_contents.getvalue()
    # Validate that another cart was not used to transport this content
    assert header == "EXTRA CART CONTENT"

    # Malpz'd files should remain as malpz
    httpx_mock.reset()
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/source\?.*"),
        method="POST",
        status_code=200,
        json=[dummy_entity],
    )

    malpz_file = malpz.wrap(b"content", "security")
    api.binaries_data.upload(
        io.BytesIO(malpz_file),
        filename="testfile",
        source_id="source1",
        security="",
        augmented_streams=[],
    )
    last_request = httpx_mock.get_request()
    body = last_request.read()
    print(body)
    assert malpz.MALPZ_HEADER in body


def test_upload_dataless(api: Api, httpx_mock: HTTPXMock, dummy_entity: dict):
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/source\/dataless\?.*"),
        method="POST",
        status_code=200,
        json=[dummy_entity],
    )
    api.binaries_data.upload_dataless(
        "6e9472fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7d0",
        source_id="source1",
        security="",
        augmented_streams=[
            binaries_data.AugmentedStream(
                label="data1", file_name="file1.exe", contents_file_path=io.BytesIO(b"type1")
            ),
            binaries_data.AugmentedStream(
                label="data2", file_name="file2.exe", contents_file_path=io.BytesIO(b"type2")
            ),
            binaries_data.AugmentedStream(
                label="data3", file_name="file3.exe", contents_file_path=io.BytesIO(b"type3")
            ),
        ],
    )
    last_request = httpx_mock.get_request()
    body = last_request.read()
    assert b'Content-Disposition: form-data; name="stream_data"; filename="file1.exe"\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_data"; filename="file2.exe"\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_data"; filename="file3.exe"\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_labels"\r\n\r\ndata1\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_labels"\r\n\r\ndata2\r\n' in body
    assert b'Content-Disposition: form-data; name="stream_labels"\r\n\r\ndata3\r\n' in body
    assert b'Content-Disposition: form-data; name="security"\r\n\r\n\r\n' in body

    # test security
    httpx_mock.reset()
    pytest.raises(
        ValueError,
        api.binaries_data.upload_dataless,
        "6e9472fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7d0",
        source_id="source1",
        security=["OFFICIAL"],
        filename="dummy",
    )

    # test bad parentID (not a sha256)
    pytest.raises(
        ValueError,
        api.binaries_data.upload_dataless,
        "fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7d0",
        source_id="source1",
        security=["OFFICIAL"],
        filename="dummy",
    )

    pytest.raises(
        ValueError,
        api.binaries_data.upload_dataless,
        "6e9472fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7dZ",
        source_id="source1",
        security=["OFFICIAL"],
        filename="dummy",
    )


def test_upload_child(api: Api, httpx_mock: HTTPXMock, dummy_entity: dict):
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/child\?.*"), method="POST", status_code=200, json=[dummy_entity]
    )
    api.binaries_data.upload_child(
        io.BytesIO(b"content"),
        "6e9472fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7d0",
        {"extracted": "Extracted by 7zip."},
        filename="child_test.txt",
        security="",
    )
    last_request = httpx_mock.get_request()
    body = last_request.read()
    assert b'filename="child_test.txt"' in body
    assert b'Content-Disposition: form-data; name="filename"\r\n' in body

    # test security
    httpx_mock.reset()
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/child\?.*"), method="POST", status_code=200, json=[dummy_entity]
    )
    api.binaries_data.upload_child(
        io.BytesIO(b"content"),
        "6e9472fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7d0",
        {"extracted": "Extracted by 7zip."},
        security="OFFICIAL",
        filename="child_test.txt",
    )
    last_request = httpx_mock.get_request()
    body = last_request.read()
    print(body.decode("utf-8", errors="replace"))
    assert b'Content-Disposition: form-data; name="security"\r\n\r\nOFFICIAL\r\n' in body

    # more security tests
    httpx_mock.reset()
    httpx_mock.add_response(
        url=re.compile(rf".*\/api\/v0\/binaries\/child\?.*"), method="POST", status_code=200, json=[dummy_entity]
    )
    api.binaries_data.upload_child(
        io.BytesIO(b"content"),
        "6e9472fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7d0",
        {"extracted": "Extracted by 7zip."},
        security="OFFICIAL//TLP:GREEN",
        filename="child_test.txt",
    )
    body = httpx_mock.get_request().read()
    print(body.decode("utf-8", errors="replace"))
    assert b'Content-Disposition: form-data; name="security"\r\n\r\nOFFICIAL//TLP:GREEN\r\n' in body

    # test no filename
    pytest.raises(
        ValueError,
        api.binaries_data.upload_child,
        io.BytesIO(b"content"),
        "6e9472fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7d0",
        {"extracted": "Extracted by 7zip."},
        security="OFFICIAL",
    )

    # test bad parentID
    pytest.raises(
        ValueError,
        api.binaries_data.upload_child,
        io.BytesIO(b"content"),
        "ze9472fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7d0",
        {"extracted": "Extracted by 7zip."},
        security="OFFICIAL",
        filename="child_test.txt",
    )

    pytest.raises(
        ValueError,
        api.binaries_data.upload_child,
        io.BytesIO(b"content"),
        "16641974ddcd6ca556ce7d0",
        {"extracted": "Extracted by 7zip."},
        security="OFFICIAL",
        filename="child_test.txt",
    )

    # Test bad reference
    pytest.raises(
        ValueError,
        api.binaries_data.upload_child,
        io.BytesIO(b"content"),
        "6e9472fa56a71791f051613a0198a8c15948136e916641974ddcd6ca556ce7d0",
        {},
        security="OFFICIAL",
        filename="child_test.txt",
    )


def test_download_bulk(api: Api, httpx_mock: HTTPXMock, dummy_entity: dict):
    httpx_mock.add_response(
        url=f"{api.config.azul_url}/api/v0/binaries/content/bulk", method="POST", status_code=200, json=[dummy_entity]
    )

    api.binaries_data.download_bulk(
        [
            "52b526411070a0a92075ea7c2575f759f480f2f4788d56300091696fc7eabb71a74a5fbca04b1934e215ca00bb6b977f6069a34588caa81f622616caacbc83bf",
        ]
    )
    body = httpx_mock.get_request().read()
    print(body.decode("utf-8", errors="replace"))
    data = json.loads(body.decode("utf-8", errors="replace"))
    assert (
        "52b526411070a0a92075ea7c2575f759f480f2f4788d56300091696fc7eabb71a74a5fbca04b1934e215ca00bb6b977f6069a34588caa81f622616caacbc83bf"
        == data["binaries"][0]
    )


def test_get_meta(api: Api, httpx_mock: HTTPXMock, dummy_entity_meta: dict):
    dummy_meta = dummy_entity_meta
    httpx_mock.add_response(
        url=re.compile(rf"{api.config.azul_url}/api/v0/binaries/[A-Za-z0-9]{{128}}"),
        method="GET",
        status_code=200,
        json={"data": dummy_meta},
    )

    api.binaries_meta.get_meta(
        "52b526411070a0a92075ea7c2575f759f480f2f4788d56300091696fc7eabb71a74a5fbca04b1934e215ca00bb6b977f6069a34588caa81f622616caacbc83bf"
    )
    path = httpx_mock.get_request().url.path
    assert (
        "52b526411070a0a92075ea7c2575f759f480f2f4788d56300091696fc7eabb71a74a5fbca04b1934e215ca00bb6b977f6069a34588caa81f622616caacbc83bf"
        in path
    )


def _setup_find_mock(api: Api, httpx_mock: HTTPXMock):
    httpx_mock.reset()
    httpx_mock.add_response(
        url=re.compile(rf"{api.config.azul_url}/api/v0/binaries.*"),
        method="POST",
        status_code=200,
        json={"data": models_restapi.EntityFind(items_count=0, items=[]).model_dump()},
    )


def _make_find_assertions(
    httpx_mock: HTTPXMock,
    term: str = [],
    max_entities: int = [],
    sort_prop: models_restapi.FindBinariesSortEnum = [],
    sort_asc: bool = [],
):
    req = httpx_mock.get_requests()
    assert req[0].url.params.get_list("term") == term
    assert req[0].url.params.get_list("max_entities") == max_entities
    assert req[0].url.params.get_list("sort") == sort_prop
    assert req[0].url.params.get_list("sort_asc") == sort_asc


def test_find(api: Api, httpx_mock: HTTPXMock):
    _setup_find_mock(api, httpx_mock)
    api.binaries_meta.find(term="source.name:testing")
    _make_find_assertions(httpx_mock, ["source.name:testing"])

    _setup_find_mock(api, httpx_mock)
    api.binaries_meta.find(
        term="source.name:testing",
        max_entities=10,
        sort_prop=models_restapi.FindBinariesSortEnum.timestamp,
        sort_asc=True,
    )
    _make_find_assertions(
        httpx_mock,
        ["source.name:testing"],
        max_entities=[str(10)],
        sort_prop=[models_restapi.FindBinariesSortEnum.timestamp],
        sort_asc=["true"],
    )

    _setup_find_mock(api, httpx_mock)
    api.binaries_meta.find(
        term="source.name:testing", sort_prop=models_restapi.FindBinariesSortEnum.source_timestamp, sort_asc=False
    )
    _make_find_assertions(
        httpx_mock,
        ["source.name:testing"],
        max_entities=[],
        sort_prop=[models_restapi.FindBinariesSortEnum.source_timestamp],
        sort_asc=["false"],
    )



"""Seed data into Azul to ensure it has the expected binaries already in it before starting any test queries."""

import io
import random
import string
import time

from azul_client import config
from azul_client.api import Api

__all__ = ["module_sha256s", "module_source", "module_ref"]


module_sha256s: list[str] = []
module_source: str = "testing"
module_ref: dict = {"user": "azul-client-test"}
# must provide a valid security string to server during file uploads
upload_security: str = "OFFICIAL"


def setUpModule():
    """Upload all of the assumed to be present files to azul and then wait until they are queryable in azul.

    Max wait is 100 seconds so if it takes
    """
    print("setting up module")
    api = Api(config.get_config())
    temp_kwargs = {
        "source_id": module_source,
        "security": upload_security,
        "references": module_ref,
        "refresh": True,  # Make sure parent binaries are in the system ASAP so we can use them for tests
    }
    number_of_binaries = 10  # Number of binaries to generate
    for i in range(number_of_binaries):
        # Random content to make each test run unique.
        random_string = "".join(random.choices(string.ascii_letters + string.digits, k=500))
        random_content = random_string.encode()
        # x 100 to get size
        std_content = io.BytesIO(
            b"content of generic files uploaded to azul for client tests at least 4096 bytes " * 100
            + b"worth for ssdeep and tlsh etc"
            + bytes("-postfix-" + str(i), "utf-8")
            + random_content
        )

        temp_kwargs["filename"] = f"test_basic_upload-{i}.txt"
        entity = api.binaries_data.upload(std_content, **temp_kwargs)
        module_sha256s.append(entity.sha256)
    print(f"Module file hashes: {module_sha256s}")  # run test with `pytest -s` to see this print out

    start_time = time.time()
    max_wait = 100  # wait up to 100 seconds for binaries to be in the system.
    all_binaries_found = False
    while start_time + max_wait > time.time():
        print("Checking if the module hashes are ready to be queried against for tests.")
        missing = False
        for sha in module_sha256s:
            if not api.binaries_meta.check_meta(sha):
                missing = True
            if not api.binaries_data.check_data(sha):
                missing = True
        if not missing:
            # No binaries are missing they're all queryable in azul now so stop waiting.
            all_binaries_found = True
            break
        time.sleep(2)

    if not all_binaries_found:
        raise Exception(
            f"Failed to verify all files were uploaded to azul in {max_wait} seconds.\n !!! - Tests cannot run as module level seeding failed."
        )



"""Handle and register api exceptions designed to allow for more explicit message types to be added overtime."""

import httpx


class BadResponse(Exception):
    """The client received a bad http status code from the server."""

    def __init__(self, resp: httpx.Response, *args):
        message = f"{resp.url} - {resp.status_code} - {resp.content}"
        super().__init__(message, *args)


class BadResponse404(Exception):
    """The client received a bad http status code from the server."""

    def __init__(self, resp: httpx.Response, *args):
        message = f"{resp.url} - {resp.status_code} - {resp.content}"
        super().__init__(message, *args)


def bad_response(resp: httpx.Response) -> BadResponse | BadResponse404:
    """Raise a formatted exception based on an http response's status code.

    :return BadResponse | BadResponse404
    """
    # If you want a different status code e.g 500 to be explicitly raised this makes it easy to add later.
    if resp.status_code == 404:
        return BadResponse404(resp)
    return BadResponse(resp)



import os
import tempfile
from unittest import TestCase

from azul_client import config as mconfig


class BaseConfigTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        # Delete all environment variables
        cls.original_env = dict(os.environ)
        for env_key in cls.original_env.keys():
            del os.environ[env_key]

    @classmethod
    def tearDownClass(cls):
        # Set all environment variables back to what they were.
        for env_key, env_value in cls.original_env.items():
            os.environ[env_key] = env_value


class TestClient(BaseConfigTestCase):
    def tearDown(self):
        if os.environ.get("AZUL_CONFIG_LOCATION"):
            del os.environ["AZUL_CONFIG_LOCATION"]

    def test_get(self):
        with tempfile.NamedTemporaryFile("w") as f:
            f.write("""
[default]
azul_url = http://lemon
            """)
            f.seek(0)
            print(f.name)
            os.environ["AZUL_CONFIG_LOCATION"] = f.name
            conf = mconfig.get_config()

        self.assertEqual(conf.azul_url, "http://lemon")
        self.assertEqual(conf.auth_type, "callback")

        with tempfile.NamedTemporaryFile("w") as f:
            f.write("""
[default]
azul_url = http://lemon/
            """)
            f.seek(0)
            print(f.name)
            os.environ["AZUL_CONFIG_LOCATION"] = f.name
            conf = mconfig.get_config()

        self.assertEqual(conf.azul_url, "http://lemon")
        self.assertEqual(conf.auth_type, "callback")

        with tempfile.NamedTemporaryFile("w") as f:
            f.write("""
[default]
azul_url = http://lemon
oidc_url = here
auth_type = blah
auth_client_id = there
            """)
            f.seek(0)
            print(f.name)
            os.environ["AZUL_CONFIG_LOCATION"] = f.name
            conf = mconfig.get_config()

        self.assertEqual(conf.azul_url, "http://lemon")
        self.assertEqual(conf.oidc_url, "here")
        self.assertEqual(conf.auth_type, "blah")
        self.assertEqual(conf.auth_client_id, "there")


class TestLocation(TestCase):
    def setUp(self):
        if os.environ.get("AZUL_CONFIG_LOCATION"):
            del os.environ["AZUL_CONFIG_LOCATION"]

    def tearDown(self):
        if os.environ.get("AZUL_CONFIG_LOCATION"):
            del os.environ["AZUL_CONFIG_LOCATION"]

    def test_location_default(self):
        self.assertTrue(
            mconfig.ConfigLocation().azul_config_location.endswith(".azul.ini"),
            msg=f"{mconfig.ConfigLocation().azul_config_location} does not end with .azul.ini",
        )
        os.environ["AZUL_CONFIG_LOCATION"] = "/tmp/azul.ini"
        self.assertEqual(mconfig.ConfigLocation().azul_config_location, "/tmp/azul.ini")
